#include <math.h>

#include "geometry.h"

float point_point_distance(point_t p1, point_t p2)
{
    float dx, dy;

    dx = p2.x - p1.x;
    dy = p2.y - p1.y;

    return sqrtf(dx * dx + dy * dy);
}

float point_seg_distance(point_t pt, line_t seg)
{
    float dx, dy, ox, oy, d, dmagsq;

    dx = seg.p[1].x - seg.p[0].x;
    dy = seg.p[1].y - seg.p[0].y;
    dmagsq = dx * dx + dy * dy;

    ox = pt.x - seg.p[0].x;
    oy = pt.y - seg.p[0].y;

    /* Find the distance of <pt> in the direction of p1->p0. */
    d = dx * ox + dy * oy;

    if (d < 0)
    {
        /* Nearest point_t on the segment is p0 */
        return sqrtf(ox * ox + oy * oy);
    } else if (d > dmagsq)
    {
        /* Nearest point_t on the segment is p1 */
        return point_point_distance(pt, seg.p[1]);
    } else {
        return fabs(dx * oy - dy * ox) / sqrtf(dmagsq);
    }
}

void clip_velocity_to_line(line_t line, point_t v, point_t *out)
{
    point_t ret;
    float dx, dy, mag, t;

    /* Get the unit direction vector of the line */
    dx = line.p[1].x - line.p[0].x;
    dy = line.p[1].y - line.p[0].y;
    mag = sqrtf(dx * dx + dy * dy);
    dx /= mag;
    dy /= mag;

    /* Get the magnitude of the velocity in the direction parallel to the line */
    t = v.x * dx + v.y * dy;

    /* The new velocity is the parallel component of the original velocity */
    ret.x = t * dx;
    ret.y = t * dy;

    if (out)
    {
        *out = ret;
    }
}

int line_line_intersection(line_t l1, line_t l2, point_t *i)
{
    float d, x1, y1, x2, y2, x3, y3, x4, y4;

    x1 = l1.p[0].x;
    y1 = l1.p[0].y;
    x2 = l1.p[1].x;
    y2 = l1.p[1].y;
    x3 = l2.p[0].x;
    y3 = l2.p[0].y;
    x4 = l2.p[1].x;
    y4 = l2.p[1].y;

    // http://mathworld.wolfram.com/Line-LineIntersection.html
    d = DET(x1 - x2, y1 - y2, x3 - x4, y3 - y4);
    if (fabs(d) < 0.00001f)
        return 0;

    i->x = DET(DET(x1, y1, x2, y2), x1 - x2, DET(x3, y3, x4, y4), x3 - x4) / d;
    i->y = DET(DET(x1, y1, x2, y2), y1 - y2, DET(x3, y3, x4, y4), y3 - y4) / d;

    return 1;
}

int circle_line_intersection(line_t line, circle_t circle, point_t *pt)
{
    float x1, y1, x2, y2, r;
    float dx, dy, dr, d, disc, s, sd, dr2;

    // From MathWorld
    r = circle.radius;
    x1 = line.p[0].x - circle.center.x;
    y1 = line.p[0].y - circle.center.y;
    x2 = line.p[1].x - circle.center.x;
    y2 = line.p[1].y - circle.center.y;

    dx = x2 - x1;
    dy = y2 - y1;
    dr = sqrtf(dx * dx + dy * dy);
    d = DET(x1, x2, y1, y2);
    disc = r * r * dr * dr - d * d;

    if (dr == 0)
    {
        return 0;
    }

    if (disc < 0)
    {
        /* No solutions */
        return 0;
    }

    if (pt)
    {
        s = (dy < 0) ? -1 : 1;
        sd = sqrtf(disc);
        dr2 = dr * dr;
        pt[0].x = (d * dy + s * dx * sd) / dr2 + circle.center.x;
        pt[0].y = (-d * dx + fabs(dy) * sd) / dr2 + circle.center.y;
    }

    if (disc == 0)
    {
        /* Only one solution, so don't bother calculating the other one */
        return 1;
    }

    if (pt)
    {
        pt[1].x = (d * dy - s * dx * sd) / dr2 + circle.center.x;
        pt[1].y = (-d * dx - fabs(dy) * sd) / dr2 + circle.center.y;
    }

    return 2;
}
