#include <stdio.h>
#include <math.h>
#include <allegro.h>

#include "world.h"
#include "robot.h"
#include "win_support.h"

void draw_radius(BITMAP *buf, int cx, int cy, float angle, float r, unsigned int color)
{
    float a;

    a = angle * M_PI / 180.0f;
    line(buf, cx, cy, cx + cos(a) * r, cy + sin(a) * r, color);
}

void draw_robot(BITMAP *buf, robot_t *robot)
{
    int cx, cy, ax, ay;
    unsigned int color;
    float r = robot->circle.radius;
    float a;
    float a2;

    cx = robot->circle.center.x;
    cy = robot->circle.center.y;

    /* Robot's angle in radians */
    a = robot->angle * M_PI / 180.0f;

    /* Current direction arrow */
    ax = cx + cos(a) * r;
    ay = cy + sin(a) * r;

    a2 = (robot->angle + 10) * M_PI / 180.0f;
    line(buf, ax, ay, cx + cos(a2) * r * 0.75f, cy + sin(a2) * r * 0.75f, 0xffffff);
    a2 = (robot->angle - 10) * M_PI / 180.0f;
    line(buf, ax, ay, cx + cos(a2) * r * 0.75f, cy + sin(a2) * r * 0.75f, 0xffffff);

    line(buf, cx, cy, ax, ay, 0xcccccc);

#if 0   /* Sorry, I don't have time to make the sensor work this week */
    /* Sensor FOV */
    if (robot->sees_obstacle)
    {
        color = 0xff0000;    /* Red */
    } else {
        color = 0x0000ff;    /* Blue */
    }

    draw_radius(buf, cx, cy, robot->angle + 45, r, color);
    draw_radius(buf, cx, cy, robot->angle - 45, r, color);
#endif

    /* Body */
    if (robot->collides)
    {
        color = 0xff0000;   /* Red */
    } else {
        color = 0x0000ff;   /* Blue */
    }

    circle(buf, cx, cy, robot->circle.radius, color);
}

unsigned int waypoint_color(int num)
{
    if (num <= 0)
    {
        /* Not a waypoint */
        return 0xff00;
    } else if (num <= touched_waypoint)
    {
        /* A waypoint that has been touched */
        return 0xffffff;
    } else {
        /* A waypoint that has not been touched */
        return 0x4040ff;
    }
}

void draw_walls(BITMAP *buf)
{
    wall_t *wall;

    for (wall = first_wall; wall; wall = wall->next)
    {
        line(buf, wall->line.p[0].x, wall->line.p[0].y, wall->line.p[1].x, wall->line.p[1].y, waypoint_color(wall->waypoint));
    }
}

void draw_waypoints(BITMAP *buf)
{
    waypoint_t *wpt;

    for (wpt = first_waypoint; wpt; wpt = wpt->next)
    {
        unsigned int color = waypoint_color(wpt->num);
        int x = wpt->circle.center.x;
        int y = wpt->circle.center.y;
        int r = wpt->circle.radius;

        circle(buf, x, y, r, color);
        textprintf_centre_ex(buf, font, x, y, color, 0, "%d", wpt->num);
    }
}

int main(int argc, char *argv[])
{
    BITMAP *buf;
    robot_t robot;
    const char *filename;
    char name[256];
    unsigned int old_ticks, ticks, ms;
    float dtime;

    if (argc == 2)
    {
        filename = argv[1];
    } else {
        if (!get_file_name(name, sizeof(name)))
            return 1;
        filename = name;
    }

    if (!world_load(filename))
    {
        printf("Can't load %s\n", filename);
        return 1;
    }

    /* Set up the robot */
    robot_init(&robot);
    robot.circle.center.x = 16;
    robot.circle.center.y = 16;
    robot.circle.radius = 15;
    robot.angle = 0;

    /* Set up Allegro */
    allegro_init();
    install_keyboard();

    set_color_depth(32);
    if (set_gfx_mode(GFX_AUTODETECT, 600, 400, 0, 0))
    {
        printf("Failed to set graphics mode\n");
        return 1;
    }

    buf = create_bitmap(SCREEN_W, SCREEN_H);

    /* Main loop */
    old_ticks = get_time();
    while (!keypressed())
    {
        /* Time since last iteration */
        ticks = get_time();
        dtime = (ticks - old_ticks) / 1000.0f;
        old_ticks = ticks;

        /* Interface to the robot control code */
        robot_current_velocity = robot.velocity;
        robot_current_angle = robot.angle;
        robot_hit = robot.collides;
        robot_sense = robot.sees_obstacle;
        robot_run();
        robot.set_velocity = robot_forward;
        robot.set_turn = robot_turn;

        /* Movement happens here */
        robot_update(&robot, dtime);

        /* Drawing starts here*/
        clear_bitmap(buf);

        /* Status text */
        textprintf_ex(buf, font, 5, 5, 0xcccccc, 0, "Waypoint %d", touched_waypoint);
        if (touched_waypoint == last_waypoint)
            textout_ex(buf, font, "Done!", 5, 15, 0xffffff, 0);
        textprintf_ex(buf, font, 5, 25, 0xcccccc, 0, "State %d", robot_state);

        draw_walls(buf);
        draw_waypoints(buf);
        draw_robot(buf, &robot);

        vsync();    /* Does this actually do anything on Windows? */
        blit(buf, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);

        /* Limit maximum framerate */
        ms = get_time() - old_ticks;
        if (ms < 20)
        {
            msleep(20 - ms);
        }
    }

    return 0;
}
END_OF_MAIN()
