#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>

#include "world.h"

/* This really shouldn't be here, but I need to set robot_waypoint and it's easiest to do it from this file */
#include "robot.h"

int last_waypoint = 0;
int touched_waypoint = 0;

wall_t *first_wall = NULL;
waypoint_t *first_waypoint = NULL;

void wall_add(wall_t *wall)
{
    wall->next = first_wall;
    first_wall = wall;
}

void waypoint_add(waypoint_t *wpt)
{
    wpt->next = first_waypoint;
    first_waypoint = wpt;
}

/* Called when the robot touches a waypoint circle or wall */
void waypoint_touch(int num)
{
    if (num == (touched_waypoint + 1))
    {
        touched_waypoint = num;
    }
}

/* Special case handler for move_to_wall where one of the wall's endpoints is inside or on the circle. */
static void move_to_wall_endpoint(circle_t *circle, point_t v, point_t endpoint, line_t l1)
{
    point_t pt[2];
    int n, i;
    float d[2];
    float x1, y1;
    circle_t tc;

    /* Move to the first endpoint */

    /* Starting point, for convenience */
    x1 = circle->center.x;
    y1 = circle->center.y;

    tc.center = endpoint;
    tc.radius = circle->radius;
    n = circle_line_intersection(l1, tc, pt);
    if (n == 0)
    {
        return;
    }

    for (i = 0; i < n; i++)
    {
        d[i] = (pt[i].x - x1) * v.x + (pt[i].y - y1) * v.y;
    }

    if (n == 1 || d[0] < d[1])
    {
        circle->center = pt[0];
    } else {
        circle->center = pt[1];
    }
}

/* Determines if the robot is touching the end (rather than the side) of a wall.
 * Used by move_to_wall.
 */
static int use_endpoint(point_t p1, point_t p2, point_t c)
{
    float wx, wy, cx, cy;

    /* Direction of wall, going out of circle
     * (Don't make a wall that fits entirely in the robot)
     */
    wx = p2.x - p1.x;
    wy = p2.y - p1.y;

    /* Direction from the endpoint inside the circle to the circle's center */
    cx = c.x - p1.x;
    cy = c.y - p1.y;

    return (cx * wx + cy * wy) < 0;
}

/* Given a circle, its direction vector, and a wall, moves it along the direction vector to the point where it just touches the wall.
 * Makes no change if v is parallel to w.
 * The new point is guaranteed to be on the line containing the original point and in the direction of v.
 */
void move_to_wall(circle_t *circle, point_t v, line_t w)
{
    float d;
    float nx, ny, nm;
    line_t l1, l2;

    /* Line containing the point's movement */
    l1.p[0].x = circle->center.x;
    l1.p[0].y = circle->center.y;
    l1.p[1].x = circle->center.x + v.x;
    l1.p[1].y = circle->center.y + v.y;

    if (point_point_distance(circle->center, w.p[0]) <= circle->radius && use_endpoint(w.p[0], w.p[1], circle->center))
    {
        move_to_wall_endpoint(circle, v, w.p[0], l1);
        return;
    } else if (point_point_distance(circle->center, w.p[1]) <= circle->radius && use_endpoint(w.p[1], w.p[0], circle->center))
    {
        move_to_wall_endpoint(circle, v, w.p[1], l1);
        return;
    }

    /* Vector perpendicular to the wall */
    nx = w.p[1].y - w.p[0].y;
    ny = w.p[0].x - w.p[1].x;

    /* Change n's magnitude to r */
    nm = sqrtf(nx * nx + ny * ny);
    if (nm == 0)
    {
        return;
    }
    nx = nx * circle->radius / nm;
    ny = ny * circle->radius / nm;

    /* Make sure n is on the same side of the wall as the original position */
    d = (nx * v.x + ny * v.y);
    if (d == 0)
    {
        /* Direction is parallel to the wall, so we can't move */
        return;
    }
    if (d > 0)
    {
        nx = -nx;
        ny = -ny;
    }

    /* Line parallel to the wall and the circle's radius away from it, on the side of the circle's original center */
    l2.p[0].x = w.p[0].x + nx;
    l2.p[0].y = w.p[0].y + ny;
    l2.p[1].x = w.p[1].x + nx;
    l2.p[1].y = w.p[1].y + ny;

    line_line_intersection(l1, l2, &circle->center);
}

void robot_init(robot_t *robot)
{
    memset(robot, 0, sizeof(robot_t));
}

/* Changes *cur by limit or (*cur - set) * dtime, whichever is smaller. */
void rate_limit(float *cur, float set, float dtime, float limit)
{
    float delta;

    delta = set - *cur;
    if (delta < (-limit * dtime))
    {
        delta = -limit * dtime;
    } else if (delta > (limit * dtime))
    {
        delta = limit * dtime;
    }

    *cur += delta;
}

void robot_update(robot_t *robot, float dtime)
{
    point_t v, origV, new_pos;
    wall_t *wall;
    waypoint_t *wpt;

    robot_waypoint = 0;

#if 1
    /* Simple motor model */
    rate_limit(&robot->velocity, robot->set_velocity, dtime, 250);
    rate_limit(&robot->turn, robot->set_turn, dtime, 250);
#else
    robot->velocity = robot->set_velocity;
    robot->turn = robot->set_turn;
#endif

    robot->angle += robot->turn * dtime;

    /* Keep the angle in the range [0, 360). */
    if (robot->angle < 0)
    {
        robot->angle += 360;
    } else if (robot->angle >= 360)
    {
        robot->angle -= 360;
    }

    /* Calculate velocity vector from angle and velocity in that direction */
    v.x = robot->velocity * cos(robot->angle * M_PI / 180.0f);
    v.y = robot->velocity * sin(robot->angle * M_PI / 180.0f);
    origV = v;

    /* Calculate the new position.
     * This will be modified below if the robot hits something.
     */
    new_pos.x = robot->circle.center.x + v.x * dtime;
    new_pos.y = robot->circle.center.y + v.y * dtime;

    /* Check for collisions with walls */
    robot->collides = 0;
    for (wall = first_wall; wall; wall = wall->next)
    {
        float dist = point_seg_distance(new_pos, wall->line) - robot->circle.radius;
        if (dist <= 0)
        {
            if (wall->waypoint > 0)
            {
                /* Waypoints aren't solid */
                robot_waypoint = wall->waypoint;
                waypoint_touch(wall->waypoint);
            } else {
                /* Hit a solid wall */
                robot->collides = 1;

                if (dist < 0)
                {
                    circle_t circle;

                    /* Modify the velocity so it doesn't run into the wall */
                    clip_velocity_to_line(wall->line, v, &v);

                    /* Move the robot back along the original velocity to the first point at which it just touches the wall.
                     *
                     * It actually looks better if we don't do this and discard the movement for this frame when a collision happens, but
                     * I think this will handle cases that would otherwise be even worse.
                     *
                     * It looks like the robot bounces because of rounding error.  I don't have time to fix it right now.
                     */
                    circle.center = new_pos;
                    circle.radius = robot->circle.radius;
                    move_to_wall(&circle, origV, wall->line);

#if 1
                    printf("%.3f, %.3f -> %.3f, %.3f         \r", new_pos.x, new_pos.y, circle.center.x, circle.center.y);
                    fflush(stdout);
                    {
                    float d = point_point_distance(new_pos, circle.center);
                    if (d > 5)
                    {
                        printf("\n");
                    }
                    if (d > robot->circle.radius || isnan(new_pos.x) || isnan(new_pos.y))
                    {
                        printf("AHHHHH!\n");
                        exit(1);
                    }
                    }
#endif

                    new_pos = circle.center;
                }
            }
        }
    }

    /* FIXME - Shouldn't move for the whole dtime, since part of the move was taken up by moving to the wall */
    robot->circle.center.x = new_pos.x + v.x * dtime;
    robot->circle.center.y = new_pos.y + v.y * dtime;

    /* Check for touching waypoints (other than walls) */
    for (wpt = first_waypoint; wpt; wpt = wpt->next)
    {
        if (point_point_distance(robot->circle.center, wpt->circle.center) < (robot->circle.radius + wpt->circle.radius))
        {
            robot_waypoint = wpt->num;
            waypoint_touch(wpt->num);
        }
    }

    //FIXME - Sensor
    robot->sees_obstacle = 0;
}

int world_load(const char *filename)
{
    FILE *fp;
    char buf[256], *tok;
    wall_t *wall;
    waypoint_t *wpt;

    printf("Loading %s\n", filename);

    fp = fopen(filename, "rt");
    if (!fp)
    {
        printf("%s\n", strerror(errno));
        return 0;
    }

    while (!feof(fp))
    {
        buf[0] = 0;
        fgets(buf, sizeof(buf), fp);

        if (buf[0] == 'L')
        {
            wall = calloc(1, sizeof(wall_t));

            tok = strtok(buf + 1, " ");
            wall->line.p[0].x = atof(tok);

            tok = strtok(NULL, " ");
            wall->line.p[0].y = atof(tok);

            tok = strtok(NULL, " ");
            wall->line.p[1].x = atof(tok);

            tok = strtok(NULL, " ");
            wall->line.p[1].y = atof(tok);

            tok = strtok(NULL, " ");
            if (tok)
            {
                wall->waypoint = atoi(tok);
                if (wall->waypoint > last_waypoint)
                {
                    last_waypoint = wall->waypoint;
                }
            } else {
                wall->waypoint = -1;
            }

            wall_add(wall);
        } else if (buf[0] == 'W')
        {
            wpt = calloc(1, sizeof(waypoint_t));

            tok = strtok(buf + 1, " ");
            wpt->circle.center.x = atof(tok);

            tok = strtok(NULL, " ");
            wpt->circle.center.y = atof(tok);

            tok = strtok(NULL, " ");
            wpt->circle.radius = atof(tok);

            tok = strtok(NULL, " ");
            wpt->num = atoi(tok);

            waypoint_add(wpt);

            if (wpt->num > last_waypoint)
            {
                last_waypoint = wpt->num;
            }
        }
    }

    fclose(fp);

    return 1;
}
